import json
import re
import sys
from collections import defaultdict
from contextlib import nullcontext
from dataclasses import dataclass
from functools import cached_property
from importlib import import_module
from operator import attrgetter
from typing import Iterable, get_origin
from warnings import warn

from .action import Plan
from .cast import cast
from .core import (
    AbstractParam,
    Arguments,
    Expansion,
    Fragment,
    ParaO,
    ParaOMeta,
    Value,
    eager,
)
from .misc import PeekableIter, is_subseq


class CLIstr(str):
    __slots__ = ("empty",)

    def __new__(cls, value: str):
        self = super().__new__(cls, "" if value is None else value)
        self.empty = value is None
        return self

    def __cast_to__(self, typ, original_type):
        if ori := get_origin(typ):
            if isinstance(ori, type):  # pragma: no branch
                if issubclass(ori, (tuple, list, set, frozenset)):
                    if len(parts := self.split(",")) > 1:
                        return cast(parts, original_type)
        if self.empty and typ is bool:
            return True

        return NotImplemented


class MalformedCommandline(ValueError): ...


class ParaONotFound(LookupError): ...


class NotAParaO(ValueError): ...


class ValueMissing(ValueError): ...


class Sep(tuple[str, ...]):
    class NeedValues(RuntimeError): ...

    class Overlap(RuntimeError): ...

    def __init__(self, _):
        super().__init__()
        if not self:
            raise self.NeedValues()

    def __lshift__(self, other: "Sep") -> "Sep":
        if overlap := set(self).intersection(other):
            raise self.Overlap(f"{self} & {other} = {overlap}")
        return self.__class__(self + other)

    def __or__(self, other: "Sep") -> "Sep":
        return self.__class__(set(self).union(other))

    @cached_property
    def parts(self) -> tuple[tuple[str, ...], tuple[str, ...]]:
        return (
            tuple(e for e in self if len(e) == 1),
            tuple(e for e in self if len(e) > 1),
        )

    @cached_property
    def regex(self):
        char, frag = self.parts
        if frag:
            frag = tuple(map(re.escape, frag))
        if char:
            frag += (f"[{re.escape(''.join(char))}]",)
        return re.compile(f"(?:{'|'.join(frag)})" if len(frag) > 1 else frag[0])

    def split1(self, string: str) -> tuple[str, str | None]:
        res = self.regex.split(string, 1)
        if len(res) > 1:
            return tuple(res)
        else:
            return res[0], None

    def split(self, string: str) -> list[str]:
        return self.regex.split(string)

    def sub(self, repl: str, string: str) -> str:
        return self.regex.sub(repl, string)


@dataclass
class CLIParser:
    uscore: Sep = "-"
    chain: Sep = "."
    module: Sep = ":"
    part: Sep = ",+"

    pair: Sep = ":"
    item: Sep = ","

    flag: Sep = ";"
    value: Sep = "="

    def __post_init__(self):
        for key, typ in CLIParser.__annotations__.items():
            if typ is Sep:  # pragma: no branch
                setattr(self, key, Sep(getattr(self, key)))

        outer = self.flag | self.value
        self._flag_value_disjoint = set(self.flag).isdisjoint(self.value)

        self.uscore << self.chain << self.module << self.part << outer
        self.pair << self.item << outer

    def argument(self, raw: str):
        if self._flag_value_disjoint:
            raw, value = self.value.split1(raw)
            raw, flags = self.flag.split1(raw)
        else:
            raw, flags = self.flag.split1(raw)
            if flags is None:
                value = None
            else:
                flags, value = self.value.split1(flags)

        if flags is not None:
            flags = dict(map(self.pair.split1, self.item.split(flags)))

        key = list(
            filter(
                any,
                map(
                    self.element,
                    self.part.split(self.uscore.sub("_", raw)),
                ),
            )
        )

        return key, flags, value

    def element(self, raw: str):
        mod, att = self.module.split1(raw)
        return (
            self.chain.sub(".", mod),
            self.chain.sub(".", att) if att else att,
        )


class NoAttributeBoundary(RuntimeWarning): ...


class UnsupportedKeyType(RuntimeWarning): ...


class MultipleCandidates(RuntimeWarning): ...


class AmbigouusCandidate(RuntimeWarning): ...


class UnusedCLIArguments(RuntimeWarning): ...


class CLI:
    prase_raw = CLIParser()

    def __init__(self, entry_points: Iterable[ParaOMeta] | None = None):
        seen = set()
        queue: list[type] = [ParaO] if entry_points is None else list(entry_points)
        for curr in queue:
            queue.extend(
                cand
                for cand in reversed(curr.__subclasses__())
                if cand.__name__[0] != "_"
                and (cand.__module__[0] != "_" or cand.__module__ == "__main__")
                and cand not in seen
            )
            seen.add(curr)

        self._paraos = seen

    @cached_property
    def find_parao(self):
        lut = defaultdict(dict)
        for s in self._paraos:
            qn = s.__qualname__.split(".")
            for i in range(len(qn)):
                sub = lut[".".join(qn[i:])]
                k = tuple(s.__module__.split("."))
                sub[k] = False if k in sub else s

        def func(module: str, attr: str) -> ParaOMeta | None:
            if sub := lut.get(attr, None):
                if module:
                    want = module.split(".")
                    cand = []

                    for have, parao in sub.items():
                        if is_subseq(want, have):
                            cand.append(parao)
                else:
                    cand = sub.values()

                if num := len(cand):
                    if num > 1:
                        w = MultipleCandidates
                    else:
                        if ret := next(iter(cand)):
                            return ret
                        else:
                            w = AmbigouusCandidate
                    warn(f"{module}:{attr}" if module else attr, w)

        return func

    def _split_case(self, raw: str):
        parts = raw.split(".")
        if len(parts) == 1:
            if raw[0].isupper():
                return "", raw, ""
        else:
            upper = [p[0].isupper() for p in parts]
            try:
                b = upper.index(True)
            except ValueError:
                pass
            else:
                upper.append(False)  # easiert than handling the -1
                e = upper.index(False, b)
                return ".".join(parts[:b]), ".".join(parts[b:e]), ".".join(parts[e:])
        return "", "", raw

    def _parse_mod_att(
        self,
        mod_att: tuple[str, str | None],
        typ: type | tuple[type],
        typ_bad: type[Warning | Exception],
    ):
        module, attr = mod_att

        if attr is None:
            module, cname, sub = self._split_case(module)
            # prepare attribute for import lookup
            attr = ".".join(filter(None, (cname, sub)))
        elif attr:
            pre, cname, sub = self._split_case(attr)
            if pre:  # some non-module prefix
                cname = ""  # skip lookup by subclass
        if attr and cname:  # lookup by subclass
            if ret := self.find_parao(module, cname):
                return attrgetter(sub)(ret) if sub else ret

        if module:
            if not attr:
                raise MalformedCommandline(f"Missing attribute for module {module}:")
            try:
                ret = attrgetter(attr)(import_module(module))
            except (ModuleNotFoundError, AttributeError) as e:
                e.add_note(f"module: {module}, attribute: {attr}")
                raise

            if not isinstance(ret, typ):  # pragma: no branch
                if issubclass(typ_bad, Warning):
                    warn(repr(ret), typ_bad)
                else:
                    raise typ_bad(repr(ret))

            return ret

        return module

    def parse_typ(self, raw: str):
        return self._parse_mod_att(self.prase_raw.element(raw), ParaOMeta, NotAParaO)

    def parse_key(self, mod_att):
        return (
            self._parse_mod_att(
                mod_att,
                (type, AbstractParam, str)[:2],  # no strings
                UnsupportedKeyType,
            )
            or mod_att[0]
        )

    def parse_args(self, args: list[str], position0: int = 100):
        pre: list[str] = []
        got: list[tuple[ParaOMeta, Arguments, list[str]]] = []

        typ: type = None
        raw: list[str] = None
        curr: list[Fragment] = None

        pit = PeekableIter(args)
        for arg in pit:
            if not arg:  # ignore empty standalone args
                continue
            if body := arg.lstrip("+-"):
                if start := arg[: -len(body)]:
                    if typ is None:
                        pre.append(arg)
                        continue

                    key, flags, value = self.prase_raw.argument(body)

                    key = tuple(map(self.parse_key, key))

                    if flags is None:
                        flags = {}

                    # fill value if it makes sense
                    if value is None:
                        if not pit.peek("+").startswith(("+", "-")):
                            value = next(pit)
                    if "class" in flags or key[-1] == "__class__":
                        if not value:
                            raise ValueMissing(arg)
                        if cls := self.parse_typ(value):
                            value = cls
                        else:
                            raise ParaONotFound(value)
                    elif "json" in flags:
                        if not value:
                            raise ValueMissing(arg)
                        try:
                            value = json.loads(value)
                        except Exception as e:
                            e.add_note(f"for argument {arg}")
                            raise
                    else:
                        value = CLIstr(value)

                    # prio
                    if prio := flags.get("prio", None):
                        try:
                            prio = int(prio)
                        except ValueError:
                            try:
                                prio = float(prio)
                            except ValueError as e:
                                e.add_note(f"for argument: {arg}")
                                raise
                    else:
                        prio = 1 - start.count("-") + start.count("+")

                    raw.append(arg)
                    curr.append(
                        Fragment.make(key, Value(value, prio, len(curr) + position0))
                    )
                else:
                    if typ is not None:
                        got.append((typ, Arguments(curr), raw))
                    # solve typ
                    typ = self.parse_typ(body)
                    if not typ:
                        raise ParaONotFound(body)
                    raw = [arg]
                    curr = []
            else:
                break

        if typ is not None:
            got.append((typ, Arguments(curr), raw))

        return pre, got, list(pit)

    def _wrap(self, pre: list[str], post: list[str]):
        if pre:
            warn(f"{pre=}", UnusedCLIArguments)
        if post:
            warn(f"{post=}", UnusedCLIArguments)

        return nullcontext(self._consume)

    def _consume(self, typ: ParaOMeta, args: Arguments, raw: list[str] = ()):
        try:
            yield from Expansion.generate(typ, args)
        except Exception as exc:
            exc.add_note(f"for arguments: {' '.join(raw)}")
            raise

    def _run(self, args: list[str]):
        pre, got, post = self.parse_args(args)

        with eager(True), self._wrap(pre, post) as consume:
            for item in got:
                yield from consume(*item)

    def run(self, args: list[str] | None = None, **kwargs):
        if args is None:
            args = sys.argv[1:]

        kwargs.setdefault("run", True)

        with Plan().use(**kwargs):
            return list(self._run(args))


if __name__ == "__main__":
    CLI().run()  # pragma: no cover



import json
import pickle
from dataclasses import KW_ONLY, dataclass
from errno import EXDEV
from io import FileIO
from pathlib import Path
from shutil import copy2, copytree, rmtree
from tempfile import NamedTemporaryFile, TemporaryDirectory
from types import GenericAlias, UnionType
from typing import Callable, Iterable, Type, TypeAliasType, _AnnotatedAlias, overload
from warnings import warn

from .core import UNSET, Unset, UntypedWarning
from .run import BaseOutput, PseudoOutput
from .shash import hex_hash

type JSON[T] = T
type Pickle[T] = T


class FSOutput(Path, PseudoOutput):
    tmpio: None | FileIO = None
    _temp: None | TemporaryDirectory = None

    def __init__(self, path, *paths):
        super().__init__(path, *paths)
        self._copy_temp(path)

    def _copy_temp(self, source):
        if isinstance(source, FSOutput):
            tmp = source._temp
            if isinstance(tmp, TemporaryDirectory) and self.is_relative_to(tmp.name):
                self._temp = tmp
        return self

    def with_segments(self, *args):
        return super().with_segments(*args)._copy_temp(self)

    @classmethod
    def tempDir(cls, **kwargs):
        tmp = TemporaryDirectory(**kwargs)
        ret = cls(tmp.name)
        ret._temp = tmp
        return ret

    def close(self):
        if (f := self.tmpio) is not None and not f.closed:
            f.close()


class Dir(FSOutput):
    @classmethod
    def temp(cls, **kwargs):
        return cls.tempDir(**kwargs)


class File(FSOutput):
    @classmethod
    def temp(cls, mode: str = "wb", **kwargs):
        tmp = NamedTemporaryFile(mode, **kwargs)
        ret = cls(tmp.name)
        ret.tmpio = tmp.file
        ret._closer = tmp._closer
        return ret


class UntypedOuput(UntypedWarning): ...


class MoveAcrossFilesystem(RuntimeWarning): ...


class MissingOuput(FileNotFoundError): ...


class NotSupported(TypeError, NotImplementedError): ...


class Incompatible(RuntimeError): ...


class Inconsistent(RuntimeWarning): ...


@dataclass
class Coder[T]:
    """En-/De-coder definition, including suffix."""

    suffix: str
    tat: TypeAliasType | None = None
    load: None | Callable[[], T] = None
    dump: None | Callable[[T], None] = None
    _: KW_ONLY
    typ: type | None = None

    @property
    def is_dir(self):
        return self.typ is not None and issubclass(self.typ, Dir)

    def match(self, hint: TypeAliasType | type):
        if isinstance(hint, TypeAliasType):
            return hint == self.tat
        elif self.typ is not None:
            if hint is None:
                return isinstance(hint, self.typ)
            return isinstance(hint, type) and issubclass(hint, self.typ)

    def conform(self, data: FSOutput) -> T:
        assert self.typ is not None
        if (have := self.is_dir) != (want := data.is_dir()):
            have = "directory" if have else "file"
            want = "directory" if want else "file"
            exc = IsADirectoryError if have else NotADirectoryError
            raise exc(f"got {have} expected a {want}")
        if not isinstance(data, self.typ):
            warn(f"got a {type(data)}, expected a {self.typ}", Inconsistent)
            data = self.typ(data)
        return data


class Output[T](BaseOutput[T]):
    """
    Basic output implementation for local file storage.
        Uses pickle by default, but also supports
        JSON and direct File and Dir output.
    """

    base = "."

    @property
    def stem(self) -> str:
        return f"{self.act.name}_{hex_hash(self.act.instance).decode('ascii')}"

    @property
    def suffix(self) -> str:
        return self.coder.suffix

    @property
    def path(self) -> Path:
        return Path(self.base, self.stem).with_suffix(self.suffix)

    @property
    def coder(self) -> Coder[T]:
        typs = tuple(self._utypes)
        if not typs:
            UntypedOuput.warn(self.act.action, self.act.instance)
            return self._coders[-1]
        for enc in self._coders:
            for typ in typs:
                if enc.match(typ):
                    return enc
        raise NotSupported(f"no coder for {typ}")

    _coders: list[Coder] = [
        Coder(".dir", typ=Dir),
        Coder(".file", typ=File),
        Coder(".json", JSON, json.load, json.dump),
        Coder(".pkl", Pickle, pickle.load, pickle.dump, typ=object),
    ]

    @property
    def _tmp_base(self) -> Path:
        """A directory to put the temp dir in, should be on the same FS to allow low cost rename."""
        return self.path.parent

    @overload
    def temp(
        self: "Output[TypeAliasType]", mode: str | None = None, **kwargs
    ) -> File: ...

    @overload
    def temp[F: FSOutput](self: "Output[F]") -> F: ...

    @overload
    def temp(self, mode: str | None = None, **kwargs) -> File: ...

    def temp(self, mode: str | None = None, **kwargs) -> FSOutput:
        path = self.path
        dps = dict(
            dir=self._tmp_base,
            prefix=path.with_suffix(".tmp").name,
            suffix=path.suffix,
        )

        if (is_dir := self.coder.is_dir) or mode == "":
            if mode or kwargs:
                raise ValueError("superfluous arguments!")
            ret = self._ftype.tempDir(**dps)
            return ret if is_dir else ret.joinpath(path.with_stem("temp").name)

        if mode is None:
            mode = "wb"

        return self._ftype.temp(mode, **dps, **kwargs)

    # implement the abstracts
    @property
    def exists(self):
        return self.path.exists()

    def remove(self, missing_ok: bool = False):
        if self.coder.is_dir:
            if not missing_ok or self.exists:
                rmtree(self.path)
                # self.path.rmdir(missing_ok)
        else:
            self.path.unlink(missing_ok)

    def load(self) -> T:
        if callable(load := self.coder.load):
            with self.path.open("rb") as f:
                return load(f)
        else:
            return self.type(self.path)

    def _temp_copy(self, data: Path) -> FSOutput:
        if self.coder.is_dir:
            return copytree(data, self.temp(), dirs_exist_ok=True)
        else:
            return copy2(data, self.temp(""))

    def _dump(self, data: FSOutput) -> FSOutput:
        if not data.exists():
            raise MissingOuput(str(data))
        data: FSOutput = self.coder.conform(data)
        data.close()
        if not (data._temp or data.tmpio):  # dont steal non-temporaries
            data = self._temp_copy(data)
            # no need to attempt EXDEV recovery via another _temp_copy
            diff_dev = False
        elif diff_dev := (data.stat().st_dev != self._tmp_base.stat().st_dev):
            warn("may be slow or fail", MoveAcrossFilesystem, stacklevel=3)
        try:
            data = data.replace(self.path)
        except OSError as err:
            if err.errno == EXDEV and diff_dev:
                warn("failed, making explicit copy", MoveAcrossFilesystem, stacklevel=3)
                # try again
                data = self._temp_copy(data)
                data = data.replace(self.path)
            else:
                raise
        return data

    def dump(self, data: T | FSOutput) -> T:
        if isinstance(data, FSOutput):
            data = self._dump(data)
            if (
                isinstance(self.type, type)
                and issubclass(self.type, FSOutput)
                and isinstance(data, self.type)
            ):
                return data
            else:
                return self.load()
        elif callable(dump := self.coder.dump):
            if isinstance(data, PseudoOutput):
                raise NotSupported(f"wont handle {type(data)}")
            tmp = self.temp("wb")
            dump(data, tmp.tmpio)
            self._dump(tmp)
            return data
        else:
            raise NotSupported(f"got {type(data)}, expected {self.type}")

    # type stuff
    @property
    def _ftype(self) -> Type[FSOutput]:
        for typ in self._utypes:
            if isinstance(typ, type) and issubclass(typ, FSOutput):
                return typ
        return Dir if self.coder.is_dir else File

    @property
    def _utypes(self) -> Iterable[type]:
        """unwrap type(s) to get a isinstance-able base"""
        queue = [self.type]
        while queue:
            typ = queue.pop()
            if typ is UNSET:
                continue
            while isinstance(typ, (GenericAlias, _AnnotatedAlias)):
                typ = typ.__origin__
            if isinstance(typ, UnionType):
                queue.extend(typ.__args__)
            else:
                yield typ

    @property
    def type(self) -> type | Unset:  # this bricks "type" in here
        action = self.act.action
        return getattr(action.func, "__annotations__", {}).get(
            "return", getattr(action, "return_type", UNSET)
        )



from typing import Any
from unittest import TestCase
from unittest.mock import Mock, call

from parao.action import (
    SimpleAction,
    Plan,
    RecursiveAction,
    ValueAction,
)
from parao.core import ParaO, Param, Prop, UntypedParameter, Value


class TestAction(TestCase):
    def test_simple(self):
        func = Mock()

        class Foo(ParaO):
            act = SimpleAction(func)

        with Plan().use(run=True):
            foo = Foo()
        func.assert_not_called()

        with Plan().use(run=True):
            foo = Foo({"act": False})
        func.assert_not_called()

        with Plan().use(run=False):
            foo = Foo({"act": True})
        func.assert_not_called()

        with Plan().use(run=True):
            foo = Foo({"act": True})
        func.assert_called_once_with(foo)

        func.reset_mock()
        foo = Foo()
        foo.act()
        func.assert_called_once_with(foo)

    def test_cov_method_1st_arg_annotation(self):
        class Bad1(ParaO):
            @ValueAction
            def bad_signature1(self): ...

        with self.assertRaises(TypeError), self.assertWarns(UntypedParameter):
            Bad1(bad_signature1=1).bad_signature1()

        class Bad2(ParaO):
            @ValueAction
            def bad_signature2(self, *, kw1, kw2): ...

        with self.assertRaises(TypeError), self.assertWarns(UntypedParameter):
            Bad2(bad_signature2=2).bad_signature2()

    def test_value_action(self):
        mock = Mock()

        class Foo(ParaO):
            @ValueAction
            def act(self, value: int):
                return mock(self, value)

            @ValueAction
            def act_no_type(self, value):
                return (123, value)

            @ValueAction[int, None]
            def act2(self, value): ...

        with self.assertRaises(TypeError):
            Foo().act()

        with Plan().use(run=True), self.assertWarns(UntypedParameter):
            foo = Foo({"act": 123})
        mock.assert_called_once_with(foo, 123)

        mock.reset_mock()
        foo.act()
        mock.assert_called_once_with(foo, 123)

        mock.reset_mock()
        mock.return_value = 456
        self.assertEqual(foo.act(321), 456)
        mock.assert_called_once_with(foo, 321)

        sentinel = object()
        self.assertEqual(foo.act_no_type(sentinel), (123, sentinel))

    def test_recursive_action(self):
        mock = Mock(return_value=None)

        class FooBase(ParaO):
            act = RecursiveAction(mock)

        class FooNoAct(FooBase):
            act = None

        class FooOtherAct(FooBase):
            act = Param[Any](None)

        class Foo1(FooBase):
            foo_no_act = Param[FooNoAct]()
            foo_other_act = Param[FooOtherAct]()

        class Foo2(FooBase):
            foo1 = Param[Foo1]()

        class Foo3(FooBase):
            foo2 = Param[Foo2]()

        class FooMix(FooBase):
            @Prop
            def inner(self) -> list[FooBase]:
                return [Foo1(), Foo2()][::-1]

            @SimpleAction
            def interrupter(self): ...

        mock.reset_mock()
        with Plan().use(run=True):
            mix4 = FooMix({"act": 3, ("inner", "act"): True})
        mock.assert_has_calls(
            [
                call(mix4, 0),
                call(mix4.inner[0], 1),
                call(mix4.inner[0].foo1, 2),
                call(mix4.inner[1], 1),
            ]
        )
        self.assertEqual(mock.call_count, 4)

        mock.reset_mock()
        with Plan().use(run=True):
            mix6 = FooMix(
                {
                    "act": 3,
                    "interrupter": Value(True, position=1),
                    (Foo2, "act"): Value(True, position=2),
                }
            )
        mock.assert_has_calls(
            [
                call(mix6, 0),
                call(mix6.inner[0], 1),
                call(mix6.inner[0].foo1, 2),
                call(mix6.inner[1], 1),
                call(mix6.inner[0], 0),
                call(mix6.inner[0].foo1, 1),
            ]
        )
        self.assertEqual(mock.call_count, 6)

        with Plan().use(run=True):
            foo3 = Foo3(act=True)
        mock.assert_has_calls(
            [call(foo3, 0), call(foo3.foo2, 1), call(foo3.foo2.foo1, 2)]
        )

        mock.reset_mock()
        foo3.act(1)
        mock.assert_has_calls([call(foo3, 0), call(foo3.foo2, 1)])

        mock.reset_mock()
        (foo3 := Foo3()).act()
        mock.assert_has_calls(
            [call(foo3, 0), call(foo3.foo2, 1), call(foo3.foo2.foo1, 2)]
        )

        mock.reset_mock()
        (foo3 := Foo3({(Foo1, "act"): False})).act()
        mock.assert_has_calls([call(foo3, 0), call(foo3.foo2, 1)])

        mock.reset_mock()
        (foo3 := Foo3()).act(1)
        mock.assert_has_calls([call(foo3, 0), call(foo3.foo2, 1)])

        mock.reset_mock()
        (foo3 := Foo3()).act(0)
        mock.assert_has_calls([call(foo3, 0)])

        mock.reset_mock()
        mock.return_value = True
        (foo3 := Foo3()).act()
        mock.assert_has_calls([call(foo3, 0)])



from abc import ABCMeta
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache, partial
from itertools import count
from math import inf
from os.path import dirname
from pickle import PicklingError
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Collection,
    Generator,
    Iterable,
    Iterator,
    Mapping,
    Protocol,
    Self,
    TypeVar,
    get_type_hints,
    overload,
)
from warnings import catch_warnings, warn
from weakref import WeakKeyDictionary

from .cast import Opaque, cast
from .misc import ContextValue, safe_len
from .shash import _SHash, bin_hash

__all__ = ["UNSET", "ParaO", "Param", "Prop", "Const"]
_warn_skip = (dirname(__file__),)


class Unset(Opaque):
    def __repr__(self):
        return "<UNSET>"


UNSET = Unset()
UNSET.__shash = _SHash().coll(Unset, (), None)
UNSET.__shash__ = lambda _: UNSET.__shash

Unset.__new__ = lambda _: UNSET

_param_counter = count()


# type KeyE = type | object | str
type KeyE = str | type | AbstractParam
type KeyT = tuple[KeyE, ...]
type KeyTE = KeyT | KeyE
type TypT = type | GenericAlias
type PrioT = int | float
type Mapish[K, V] = Mapping[K, V] | Iterable[tuple[K, V]]


@dataclass(slots=True, frozen=True)
class Value[T: Any]:
    val: T
    prio: PrioT = 0
    position: int = 0

    def __hash__(self):
        return hash((self.__class__, id(self.val), self.prio, self.position))

    def __eq__(self, other):
        return (
            isinstance(other, Value)
            and self.val is other.val
            and self.prio == other.prio
            and self.position == other.position
        )

    def __or__(self, other: "Value | None"):
        return self if other is None or self.prio > other.prio else other

    def __ror__(self, other: "Value | None"):
        return self if other is None or self.prio >= other.prio else other

    def __repr__(self):
        ret = [repr(self.val)]
        if self.prio or self.position:
            ret.append(repr(self.prio))
        if self.position:
            ret.append(repr(self.position))
        return f"{self.__class__.__name__}({', '.join(ret)})"


@dataclass(slots=True, frozen=True)
class Fragment:
    param: "str | AbstractParam | None"
    types: tuple[type] | None
    inner: "Fragment | Arguments | Value"

    @classmethod
    def make(cls, key: KeyTE, value: "Value | Fragment | Arguments"):
        assert value is None or isinstance(value, (Value, Fragment, Arguments))
        if not isinstance(key, tuple):
            key = (key,)
        return cls._make(iter(key), value)

    @classmethod
    def _make(
        cls, it: Iterator[KeyE], value: "Value | Fragment | Arguments"
    ) -> "Fragment":
        types = []
        for k in it:
            if isinstance(k, type):
                types.append(k)
            elif isinstance(k, (str, AbstractParam)):
                return cls(k, tuple(types) if types else None, cls._make(it, value))
            else:
                raise TypeError(f"invaid key components: {k!r}")
        else:
            return cls(None, tuple(types), value) if types else value

    def is_type_ok(self, ref: type):
        if typs := self.types:
            return all(issubclass(ref, typ) for typ in typs)
        return True

    @property
    def effective_key(self) -> KeyT:
        assert self.param is not None
        r = (self.types or ()) + (self.param,)
        if isinstance(i := self.inner, Fragment):
            r += i.effective_key
        return r

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.param!r}, {self.types!r}, {self.inner!r})"
        )


class Arguments(tuple["Arguments | Fragment", ...]):
    @classmethod
    def make(cls, *args: "Arguments | HasArguments | dict[KeyTE, Any]", **kwargs: Any):
        return cls._make(args + (kwargs,)) if kwargs else cls._make(args)

    @classmethod
    def _make(
        cls,
        args: "tuple[Arguments | HasArguments | dict[KeyTE, Any], ...]",
        kwargs: dict[str, Any] = None,
    ):
        sub = []
        if arg := cls._ctxargs():
            sub.append(arg)

        for arg in args:
            arg = getattr(arg, "__args__", arg)
            if isinstance(arg, cls):
                if arg:
                    sub.append(arg)
            elif isinstance(arg, dict):
                if arg:
                    sub.append(cls.from_dict(arg.items()))
            else:
                raise TypeError(f"unsupported argument type: {type(arg)}")
        if kwargs:
            sub.append(cls.from_dict(kwargs))

        return cls.from_list(sub)

    @classmethod
    def from_dict(
        cls,
        k2v: Mapping[KeyTE, Any] | Iterable[tuple[KeyTE, Any]],
        prio: PrioT = 0,
    ):
        if callable(items := getattr(k2v, "items", None)):
            k2v = items()
        return cls(
            Fragment.make(k, v if isinstance(v, Value) else Value(v, prio))
            for k, v in k2v
        )

    @classmethod
    def from_list(cls, args: "list[Arguments | Fragment]") -> "Arguments":
        """Turn an iterable into arguments. Avoids unnecessary nesting or repeated creation of empty Arguments."""
        match args:
            case []:
                return cls.EMPTY
            case [Arguments()]:
                return args[0]
        return cls(args)

    def __repr__(self):
        return self.__class__.__name__ + (tuple.__repr__(self) if self else "()")

    def solve_value(self, param, owner, name) -> tuple["Arguments", "Value | None"]:
        com, val, sub = self._solve_values(owner)
        if su := sub.get(param):
            su = Arguments.from_list(su)
        else:
            su = com
        return su, val.get(param)

    @lru_cache
    def _solve_values(self, ref: "ParaOMeta"):
        com: list[Arguments | Fragment] = []
        val: dict[AbstractParam, Value] = {}
        sub: defaultdict[AbstractParam, list[Arguments | Fragment]] = defaultdict(
            com.copy
        )

        op = ref.__own_parameters__
        for arg in self:
            if isinstance(arg, Arguments):
                co, va, su = arg._solve_values(ref)
                for k, v in va.items():
                    val[k] = val.get(k) | v
                for k, v in su.items():
                    sub[k].extend(v)
                if co:
                    com.append(co)  # must be done after filling sub
            elif (k := op.got(arg.param)) and arg.is_type_ok(ref):
                if isinstance((v := arg.inner), Value):
                    val[k] = val.get(k) | v
                    if arg.types:  # do we want this?
                        sub.get(k, com).append(arg)
                else:
                    sub[k].append(v)
            else:
                com.append(arg)

        if val or sub:
            return Arguments.from_list(com), val, sub
        else:
            return self, {}, {}

    @lru_cache
    def solve_class(
        self, ref: "ParaOMeta"
    ) -> "tuple[Arguments, Value[ParaOMeta] | None]":
        sub = []
        res = res0 = Value(ref, -inf)
        tar = {None, "__class__"}
        alt = False  # sub != self

        for arg in self:
            if isinstance(arg, Arguments):
                s, r = arg.solve_class(res.val)
                if s:
                    sub.append(s)
                    alt = alt or s is not arg
                if r is None:
                    continue
            elif (
                arg.param in tar
                and isinstance((r := arg.inner), Value)
                and arg.is_type_ok(res.val)
            ):
                if arg.param:
                    alt = True
                else:
                    sub.append(arg)
            else:
                sub.append(arg)
                continue
            res |= r
            if res.val is UNSET:
                res = res0

        return self.from_list(sub) if alt else self, None if res is res0 else res

    def get_root_of(self, seek: Value) -> Fragment | None:
        for root in self:
            curr = root
            while isinstance(curr, Fragment):
                curr = curr.inner
            if curr is seek:
                return root
            elif isinstance(curr, Arguments):
                if sub := curr.get_root_of(seek):
                    return sub


Arguments.EMPTY = Arguments()
Arguments._ctxargs = ContextValue("ContextArguments", default=Arguments.EMPTY)


class HasArguments(Protocol):
    __args__: Arguments


eager = ContextValue[bool]("eager", default=False)


class OwnParameters(dict[str, "AbstractParam"]):
    __slots__ = "vals"
    vals: set["AbstractParam"]

    class CacheReset(RuntimeWarning): ...

    def __init__(self, cls: "ParaOMeta"):
        super().__init__(
            (name, param)
            for name in dir(cls)
            if not name.startswith("__")
            and isinstance((param := getattr(cls, name)), AbstractParam)
        )
        self.vals = set(self.values())

    def got(self, key: "str | AbstractParam") -> "AbstractParam | None":
        if key in self.vals or (key := self.get(key)):
            return key

    @classmethod
    def reset(cls):
        if cls.cache:
            warn(
                "partially filled __own_parameters__ cache reset",
                cls.CacheReset,
                skip_file_prefixes=_warn_skip,
            )
        cls.cache.clear()

    cache: dict["ParaOMeta", "OwnParameters"] = {}


class ParaOMeta(ABCMeta):
    @property
    def __fullname__(cls):
        return f"{cls.__module__}:{cls.__qualname__}"

    @property
    def __own_parameters__(cls) -> OwnParameters:
        if (val := OwnParameters.cache.get(cls)) is None:
            val = OwnParameters.cache[cls] = OwnParameters(cls)
        return val

    def __setattr__(cls, name, value):
        if not name.startswith("_"):
            OwnParameters.reset()
            if isinstance(value, AbstractParam):
                value.__set_name__(cls, name)
        return super().__setattr__(name, value)

    def __delattr__(cls, name):
        if not name.startswith("_"):
            OwnParameters.reset()
            if isinstance((old := getattr(cls, name, None)), AbstractParam):
                old.__set_name__(cls, None)
        return super().__delattr__(name)

    def __cast_from__(cls, value, original_type):
        if value is UNSET:
            return cls()
        if isinstance(value, cls):
            return value
        return cls(value)

    def __call__(
        cls, *args: Arguments | HasArguments | dict[KeyTE, Any], **kwargs: Any
    ) -> Self:
        arg = Arguments._make(args, kwargs)
        arg, val = arg.solve_class(cls)
        ret = cls.__new__(cls if val is None else val.val)
        ret.__args__ = arg
        ret.__init__()
        if eager():
            for name, param in ret.__class__.__own_parameters__.items():
                if param.eager:
                    getattr(ret, name)
        return ret


def get_inner_parao(value: Any):
    if isinstance(value, ParaO):
        yield value
    elif isinstance(value, Expansion):
        yield from value.expand()
    elif isinstance(value, (dict, list, tuple, set, frozenset)):
        queue = [iter((value,))]
        while queue:
            for curr in queue[-1]:
                if isinstance(curr, (list, tuple, set, frozenset)):
                    queue.append(iter(curr))
                    break
                elif isinstance(curr, dict):
                    queue.append(iter(curr.keys()))
                    queue.append(iter(curr.values()))
                    break
                elif isinstance(curr, ParaO):
                    yield curr
            else:
                queue.pop()


class ParaO(metaclass=ParaOMeta):
    __args__: Arguments  # | UNSET

    def __shash__(self, enc: _SHash) -> bytes:
        try:
            res = self.__shash
        except AttributeError:
            res = self.__shash = enc.coll(
                self.__class__,
                (
                    (name, vhash)
                    for name, param in self.__class__.__own_parameters__.items()
                    if param.significant
                    and (value := getattr(self, name)) is not param.neutral
                    and (vhash := enc(value)) != enc(param.neutral)
                ),
                enc.keyh,
                True,
            )
        return res

    def __eq__(self, other):
        return (
            isinstance(self, ParaO)
            and self.__class__ is other.__class__
            and bin_hash(self) == bin_hash(other)
        )

    def __hash__(self) -> int:
        return int.from_bytes(bin_hash(self)[:8])

    @property
    def __inner__(self):
        for name, param in self.__class__.__own_parameters__.items():
            if param.significant:
                yield from get_inner_parao(getattr(self, name))

    def __repr__(
        self,
        *,
        compact: bool | str = False,
        param: "AbstractParam" = None,
        param_name: str = None,
    ):
        if compact:
            if compact is True:
                compact = "..."
            items = [compact]
        else:
            items = [
                f"{name}={value!r}"
                for name, value, neutral in self.__rich_repr__()
                if (value is not UNSET if neutral is UNSET else value != neutral)
            ]
        ret = f"{self.__class__.__fullname__}({', '.join(items)})"
        if param is not None:
            if param_name is None:
                param_name = param._name(type(self)) or "???"
            ret += f".{param_name}={param!r}"
        return ret

    def __rich_repr__(self):
        for name, param in self.__class__.__own_parameters__.items():
            if param.significant and not isinstance(param, Const):
                if (neutral := param.neutral) is UNSET:
                    neutral = getattr(param, "default", UNSET)
                yield name, getattr(self, name), neutral


class TypedAlias(GenericAlias):
    _typevar2name = WeakKeyDictionary[TypeVar, str]()  # shadowed on instances!

    class TypedAliasMismatch(RuntimeWarning): ...

    class TypedAliasClash(TypeError): ...

    class TypedAliasRedefined(RuntimeWarning): ...

    def __init__(self, *arg, **kwargs):
        super().__init__()
        cls = self.__class__
        tv2n = cls._typevar2name
        for arg, tp in zip(self.__args__, self.__origin__.__type_params__):
            if name := tv2n.get(tp):
                if isinstance(arg, TypeVar):
                    if arg.__name__ != tp.__name__:
                        warn(f"{arg} -> {tp}", cls.TypedAliasMismatch, stacklevel=4)
                    cls.register(arg, name)

    def __call__(self, *args, **kwds):
        tv2n = self.__class__._typevar2name
        for arg, tp in zip(self.__args__, self.__origin__.__type_params__):
            if name := tv2n.get(tp):
                assert not isinstance(arg, TypeVar)  # already registered during init
                kwds.setdefault(name, arg)
        return super().__call__(*args, **kwds)

    @classmethod
    def convert(cls, ga: GenericAlias):
        return cls(ga.__origin__, ga.__args__)

    @classmethod
    def register(cls, tv: TypeVar, name: str):
        if got := cls._typevar2name.get(tv):
            if got != name:
                raise cls.TypedAliasClash(f"{tv} wants {name!r} already got {got!r}")
            else:
                warn(str(tv), cls.TypedAliasRedefined, skip_file_prefixes=_warn_skip)
        else:
            cls._typevar2name[tv] = name

    @classmethod
    def init_subclass(cls, subcls: "type[AbstractParam]"):
        for ob in reversed(subcls.__orig_bases__):
            if isinstance(ob, cls):
                for arg, tp in zip(ob.__args__, ob.__origin__.__type_params__):
                    if name := cls._typevar2name.get(tp):
                        if not isinstance(arg, TypeVar) and not hasattr(subcls, name):
                            setattr(subcls, name, arg)


class UntypedWarning(RuntimeWarning):
    @classmethod
    def warn(cls, param: "AbstractParam", instance: "ParaO", name: str | None = None):
        if name is None:
            name = param._name(type(instance))
        warn(
            f"{type(param)} {name} on {type(instance)}",
            category=cls,
            skip_file_prefixes=_warn_skip,
        )


class UntypedParameter(UntypedWarning): ...


type ExpansionFilter = (
    Collection[KeyE | Collection[KeyE]] | Callable[[Expansion, ParaO], bool]
)


class DuplicateParameter(RuntimeError): ...


@lru_cache
def _solve_name(param: "Param", icls: "ParaOMeta") -> str | None:
    lut = param._owner2name
    for cls in icls.__mro__:
        if cls in lut:
            return lut[cls]


@lru_cache
def _get_type_hints(cls: "ParaOMeta"):
    with catch_warnings(category=TypedAlias.TypedAliasRedefined, action="ignore"):
        return get_type_hints(cls)


### actual code
class AbstractParam[T]:
    significant: bool = True
    neutral: T = UNSET

    TypedAlias.register(T, "type")

    def __class_getitem__(cls, key):
        return TypedAlias.convert(super().__class_getitem__(key))

    def __init_subclass__(cls):
        super().__init_subclass__()
        TypedAlias.init_subclass(cls)

    __slots__ = ("__dict__", "_owner2name", "_id")

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self._owner2name = {}
        self._id = next(_param_counter)

    def __set_name__(self, cls, name):
        if name:
            old = self._owner2name.get(cls, None)
            if old is not None:
                raise DuplicateParameter(f"{self} on {cls} with names: {old}, {name}")
            self._owner2name[cls] = name
        else:
            del self._owner2name[cls]
        _solve_name.cache_clear()

    def _name(self, cls: "ParaOMeta"):
        return _solve_name(self, cls)

    def _type(self, cls: "ParaOMeta", name: str):
        typ = self.type
        if typ is UNSET:
            typ = _get_type_hints(cls).get(name, UNSET)
        return typ

    def _cast(self, raw, typ):
        try:
            exp = cast(raw, Expansion[typ])
        except TypeError:
            pass
        else:
            if isinstance(exp, Expansion):
                raise exp
        return cast(raw, typ)

    def _get(self, val: Value | None, name: str, instance: "ParaO") -> T:
        raw = UNSET if val is None else val.val
        typ = self._type(type(instance), name)
        if typ is UNSET:
            UntypedParameter.warn(self, instance, name)
            return raw
        return self._cast(raw, typ)

    def _collect(self, expansion: "Expansion", instance: "ParaO"):
        return bool(self.collect) and (
            self.collect(expansion, instance)
            if callable(self.collect)
            else any(map(expansion.test, self.collect))
        )

    def _solve(
        self,
        val: Value | None,
        name: str,
        instance: "ParaO",
        *args: Arguments | Fragment,
    ):
        if val and val.prio < self.min_prio:
            val = None

        try:
            with Arguments._ctxargs(Arguments.from_list(args)):
                return self._get(val, name, instance)
        except Expansion as exp:
            exp.process(self, instance, val)
            exp.make = partial(self._solve, val, name, instance, *args)
            return exp
        except Exception as exc:
            exc.add_note(
                f"param stack: {instance.__repr__(compact=True, param=self, param_name=name)}"
            )
            raise

    def __get__(self, instance: "ParaO", owner: type | None = None) -> T:
        if instance is None:
            return self
        cls = type(instance)
        name = self._name(cls)

        sub, val = instance.__args__.solve_value(self, cls, name)

        instance.__dict__[name] = raw = self._solve(val, name, instance, sub)
        return raw

    def __reduce__(self):
        for cls_name in self._owner2name.items():
            return getattr, cls_name
        raise PicklingError(f"Can't pickle a {type(self)} never used on a ParaO")

    min_prio: float = -inf
    eager: bool = True
    collect: ExpansionFilter = ()
    type: type | Unset
    type = UNSET


class Const[T](AbstractParam[T]):
    def __init__(self, value: T, **kwargs):
        super().__init__(value=value, **kwargs)

    def __get__(self, instance, owner=None) -> T:
        return self if instance is None else self.value


class AbstractDecoParam[T, F: Callable](AbstractParam[T]):
    def __init__(self, func: F, **kwargs):
        assert callable(func)
        super().__init__(func=func, **kwargs)

    @overload
    def __new__(cls: Self, func: None = None, **kwargs) -> partial[Self]: ...

    @overload
    def __new__(cls: Self, func: F = None, **kwargs) -> Self: ...

    def __new__(cls, func: F | None = None, **kwargs):
        if func is None:
            return partial(cls, **kwargs)
        return super().__new__(cls)

    def __getattr__(self, name):
        if not name.startswith("_"):
            return getattr(self.func, name)
        return super().__getattr__(name)

    func: F


class Prop[T](AbstractDecoParam[T, Callable[[ParaO], T]]):
    def _type(self, cls, name):
        typ = getattr(self.func, "__annotations__", {}).get("return", UNSET)
        if typ is UNSET:
            typ = super()._type(cls, name)
        return typ

    def _get(self, val, name, instance) -> T:
        raw = super()._get(val, name, instance)
        return self.func(instance) if raw is UNSET else raw


class MissingParameterValue(TypeError): ...


class Param[T](AbstractParam[T]):
    def __init__(self, default=UNSET, **kwargs):
        super().__init__(default=default, **kwargs)

    def _get(self, val, name, instance):
        raw = super()._get(val, name, instance)
        if raw is UNSET:
            if self.default is UNSET:
                raise MissingParameterValue(name)
            else:
                return self.default
        return raw


class ExpansionGeneratedKeyMissingParameter(RuntimeWarning): ...


class Expansion[T](BaseException):
    @classmethod
    def __cast_from__(cls, value, original_type):
        (typ,) = original_type.__args__
        result = cast(value, tuple[typ, ...])
        if isinstance(result, tuple):
            return cls(result)
        else:
            return NotImplemented

    def __init__(self, values: Iterable[T]):
        super().__init__()
        assert iter(values)  # ensure values is iterable
        self.values = values
        self._frames: list[tuple[ParaOMeta, Param, Value | None]] = []

    make: Callable[[Value], ParaO | Self] | None = None

    def test(self, item: KeyE | Iterable[KeyE]):
        match item:
            case AbstractParam():
                return self.param is item
            case str():
                return bool(self.param_name == item)
            case type():
                return isinstance(self.source, item)
            case _:
                return all(map(self.test, item))

    def process(self, param: AbstractParam, inst: ParaO, value: Value | None):
        # leave marks of origin
        if not self._frames:
            self.param = param
            self.source = inst
            # do we need these? or are they only for _unwind construction
            self.value = value
        # is it collected here?
        if param._collect(self, inst):
            return  # this will add ._get
        # keep track of key to dial-down to the origin
        self._frames.append(
            (inst.__class__, param, inst.__args__.get_root_of(self.value))
        )
        raise  # self # but don't to avoid mangling the traceback

    def make_key(
        self,
        use_arg: bool = True,
        *,
        dont: Collection[KeyE] = (),
        use_cls: bool = True,
        use_param: bool = True,
        use_name: bool = True,
        want: Collection[KeyE] | None = None,
    ):
        dont = set(dont)
        dont.difference_update(
            dont_cls := tuple(d for d in dont if isinstance(d, type))
        )

        if want is not None:
            want = set(want)
            want.difference_update(
                want_cls := tuple(w for w in want if isinstance(w, type))
            )

        rkey = []
        for cls, param, root in self._frames:
            if use_arg and root is not None:
                rkey = list(root.effective_key[::-1])
                continue

            name = _solve_name(param, cls)

            if (
                use_param
                and param not in dont
                and name not in dont
                and (want is None or param in want or name in want)
            ):
                rkey.append(name if use_name else param)
            if (
                use_cls
                and not issubclass(cls, dont_cls)
                and (want is None or issubclass(cls, want_cls))
            ):
                if not rkey:
                    rkey.append(name if use_name else param)
                    warn(
                        f"force added {rkey[0]!r}",
                        ExpansionGeneratedKeyMissingParameter,
                    )
                rkey.append(cls)
        return tuple(reversed(rkey))

    def expand(self, prio: PrioT = 0, **kwargs) -> Generator[ParaO, None, None]:
        key = self.make_key(**kwargs)
        for val in self.values:
            res = self.make(Fragment.make(key, Value(val, prio)))
            if isinstance(res, Expansion):
                yield from res.expand(prio=prio, **kwargs)
            else:
                yield res

    @staticmethod
    def generate(
        typ: ParaOMeta, args: Arguments, **kwargs
    ) -> Generator[ParaO, None, None]:
        try:
            yield typ(args)
        except Expansion as exp:
            exp.make = lambda arg: typ(Arguments((args, arg)))
            try:
                yield from exp.expand(**kwargs)
            except Exception as exc:
                exc.add_note(f"while expanding: {exp!r}")
                raise

    @property
    def param_name(self):
        return self.param._name(type(self.source))

    def __repr__(self):
        parts = [f"<{safe_len(self.values, '???')} values>"]
        if self._frames:
            parts.append(f"from {self.source.__repr__(compact=True, param=self.param)}")
        return f"{self.__class__.__name__}({' '.join(parts)})"



from binascii import b2a_hex
from collections import OrderedDict
from functools import lru_cache
from hashlib import sha256
from operator import attrgetter
from types import FunctionType, ModuleType
from typing import Any, Callable, Hashable, Iterable, Protocol
from warnings import warn

primitives = (str, int, float, type(None), bytes, complex, type(Ellipsis))

importables = (type, FunctionType, ModuleType)

try:
    from numpy import dtype, ndarray
except ModuleNotFoundError:

    class ndarray: ...  # phony placeholder

else:
    primitives += (dtype,)


primitives_set = frozenset(primitives)


class UsesLocalsError(ValueError): ...


class UsesMainWarning(RuntimeWarning): ...


class UsesPickleWarning(RuntimeWarning): ...


class UnsupportedError(ValueError): ...


class _SHash(list):
    end = sha256(b"").digest()

    def __call__(self, value: Any) -> bytes:
        cls = value.__class__
        if cls in primitives_set:
            return _repr(value)
        if cls is tuple:
            return self.coll(cls, value, self)
        if cls is frozenset:
            return self.coll(cls, value, self, True)
        if isinstance(value, importables):
            return _imp(value)

        try:
            idx = self.index(id(value))
        except ValueError:
            pass
        else:
            return sha256(f"self@{len(self) - idx}".encode()).digest()

        try:
            self.append(id(value))

            if shash := getattr(value, "__shash__", None):
                return shash(self)

            func = self
            sort = False
            if isinstance(value, ndarray):
                value = value.tolist()
            elif isinstance(value, dict):
                sort = not isinstance(value, OrderedDict)
                func = self.item
                value = value.items()
            elif isinstance(value, (list, tuple)):
                pass
            elif isinstance(value, (set, frozenset)):
                sort = True
            elif isinstance(value, (range, slice)):
                value = attrgetter("start", "stop", "step")(value)
            elif isinstance(value, primitives):
                return _repr(value)
            elif callable(red := getattr(value, "__reduce__", None)):
                warn(UsesPickleWarning(cls))
                value = red()
            else:
                raise UnsupportedError(value)

            return self.coll(cls, value, func, sort)
        finally:
            self.pop()

    def item(self, item: tuple[Hashable, Any]) -> bytes:
        h = sha256(self(item[0]))
        h.update(self(item[1]))
        return h.digest()

    def keyh(self, item: tuple[Hashable, bytes]) -> bytes:
        h = sha256(self(item[0]))
        h.update(item[1])
        return h.digest()

    def coll[T](
        self,
        cls: type,
        vals: Iterable[T],
        func: Callable[[T], bytes],
        sort: bool = False,
    ) -> bytes:
        ret = sha256(_imp(cls))
        if sort:
            vals = [func(v) for v in vals]
            vals.sort()
            for val in vals:
                ret.update(val)
        else:
            for val in vals:
                ret.update(func(val))
        ret.update(self.end)
        return ret.digest()


class SHashable(Protocol):
    def __shash__(self, shash: _SHash) -> bytes: ...


def _repr(value) -> bytes:
    return sha256(repr(value).encode()).digest()


@lru_cache
def _imp(value: type | FunctionType) -> bytes:
    if isinstance(value, ModuleType):
        rep = value.__name__
    else:
        rep = f"{value.__module__}:{value.__qualname__}"
    if "<locals>" in rep:
        raise UsesLocalsError(value)
    if rep.startswith("__main__"):
        warn(UsesMainWarning(value))
    return sha256(rep.encode()).digest()


def bin_hash(value: Any) -> bytes:
    return _SHash()(value)


def hex_hash(value: Any) -> bytes:
    return b2a_hex(bin_hash(value))



from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from io import IOBase
import os
from pathlib import Path
from stat import S_IMODE, S_IWGRP, S_IWOTH, S_IWUSR
import sys
from tempfile import TemporaryDirectory
from warnings import catch_warnings
import pytest
from unittest.mock import Mock, call, patch

from parao.action import ValueAction
from parao.cli import CLI
from parao.core import Const, OwnParameters, ParaO, Param
from parao.output import (
    JSON,
    Dir,
    FSOutput,
    File,
    Inconsistent,
    MissingOuput,
    NotSupported,
    Pickle,
    MoveAcrossFilesystem,
    UntypedOuput,
)
from parao.run import ConcurrentRunner, PseudoOutput, RunAct
from parao.task import RunAction, Task, Output, pprint


sentinel = object()


class BaseTask[T](Task[T]):
    space_waster = Param[str]("")

    class output(Output):
        base = None


class Task1(BaseTask):
    code_version = 1

    @RunAction[tuple[int, str]]
    def run(self): ...

    run.aux = sentinel


class Task2(BaseTask):
    dep = Param[Task1]()

    def run(self) -> tuple:
        return 321, self.dep.run.output.load()


@pytest.fixture
def tmpdir4BaseTask(tmpdir):
    with patch.object(BaseTask.output, "base", tmpdir):
        yield tmpdir


def test(tmpdir4BaseTask):
    return_sentinel = 123, "foo"
    mock = Mock(return_value=return_sentinel)

    with patch.object(Task1.run, "func", mock):
        assert issubclass(Task1.output, Output)

        assert Task1.run.type is RunAction.type

        assert isinstance(Task1.code_version, Const)
        assert Task1.code_version.value == 1
        assert isinstance(Task1.__own_parameters__["code_version"], Const)
        assert isinstance(Task1.run, RunAction)
        assert Task1.run.aux is sentinel

        t1 = Task1()

        assert isinstance(t1.run, RunAct)
        assert t1.run.action.return_type == tuple[int, str]

        assert not t1.run.output.path.exists()
        assert not t1.run.output.exists
        assert not t1.run.done

        assert t1.run() is return_sentinel
        mock.assert_called_once_with(t1)
        assert t1.run() == return_sentinel
        assert t1.run() is not return_sentinel  # not cached (yet!?)
        mock.assert_called_once_with(t1)  # still just once
        assert t1.run.done
        assert t1.run.output.load() == (123, "foo")
        assert t1.run.output.exists
        assert t1.run.output.path.exists()

        t2 = Task2()
        assert t2.run() == (321, (123, "foo"))
        assert t2.run.output.type is tuple

        t1.run.output.remove()
        assert not t1.run.done
        t2.run.output.remove()
        assert not t2.run.done

        assert t2.run() == (321, (123, "foo"))


# test outpus stuff
class TaskX(BaseTask):
    @RunAction
    def run(self): ...


def test_output_untyped(tmpdir4BaseTask):
    t = TaskX()
    with pytest.warns(UntypedOuput):
        t.run.output.coder


@pytest.fixture
def typedTaskX(tmpdir4BaseTask, request):
    with patch.object(TaskX.run, "return_type", request.param, create=True):
        yield TaskX()


class UnsupportedPseudoOutput(PseudoOutput): ...


@pytest.mark.parametrize(
    "typedTaskX",
    [None, str | int, JSON, JSON[list], Pickle, Pickle[list]],
    indirect=True,
)
def test_output_transparent(typedTaskX):
    tmp = typedTaskX.run.output.temp()
    assert isinstance(tmp, File)
    assert isinstance(tmp.tmpio, IOBase)
    assert tmp._temp is None


def test_output_unknown(tmpdir4BaseTask):
    run = TaskX().run
    with (
        patch.object(run.output._coders[-1], "typ", None),
        patch.object(TaskX.run.func, "__annotations__", {"return": None}),
        pytest.raises(NotSupported),
    ):
        run.output.coder

    with patch.object(TaskX.run, "func") as mock:
        TaskX.run.func.__annotations__ = {"return": UnsupportedPseudoOutput}
        mock.return_value = UnsupportedPseudoOutput()
        with pytest.raises(NotSupported):
            TaskX().run()


@contextmanager
def make_readonly(target):
    target = Path(target)
    stat = S_IMODE(target.stat().st_mode)
    target.chmod(stat & ~(S_IWUSR | S_IWGRP | S_IWOTH))
    try:
        yield
    finally:
        target.chmod(stat)


def test_output_other_temp(tmpdir4BaseTask):
    with patch.object(TaskX.run, "func") as mock:
        TaskX.run.func.__annotations__ = {"return": Dir}

        if (
            other_fs := os.environ.get(
                "TEMP2_ON_DIFFERENT_FS", os.environ.get("XDG_RUNTIME_DIR", None)
            )
        ) and os.stat(tmpdir4BaseTask).st_dev != os.stat(other_fs):  # pragma: no branch
            mock.return_value = Dir.temp(dir=other_fs)
            with (
                pytest.warns(MoveAcrossFilesystem, match="slow"),
                pytest.warns(MoveAcrossFilesystem, match="failed"),
            ):
                TaskX().run()
            TaskX().run.output.remove()
            mock.assert_called_once()
            mock.reset_mock()

        tmp = Dir.temp()
        mock.return_value = tmp.parent / tmp.name
        TaskX().run()
        TaskX().run.output.remove()
        mock.assert_called_once()
        mock.reset_mock()

        # now files
        TaskX.run.func.__annotations__ = {"return": File}

        run = TaskX().run

        tmp = run.output.temp()
        mock.return_value = tmp.parent / tmp.name
        run()
        run.output.remove()
        mock.assert_called_once()
        mock.reset_mock()

        mock.return_value = run.output.temp()
        with (
            make_readonly(tmpdir4BaseTask),
            pytest.raises(OSError),
        ):  # trigger failure in rename
            run()
        mock.assert_called_once()
        mock.reset_mock()


class TaskDir(BaseTask):
    def run(self) -> Dir:
        ret: Dir = self.run.output.temp()
        ret.joinpath("foo.bar").touch()
        return ret


def test_output_Dir_alt(tmpdir4BaseTask):
    class DirAlt(Dir): ...

    with patch.object(TaskDir.run, "func") as mock:
        TaskDir.run.func.__annotations__ = {"return": DirAlt}
        mock.return_value = Dir.temp()
        TaskDir().run()
        mock.assert_called_once()


def test_output_Dir_bad(tmpdir4BaseTask):
    with patch.object(TaskDir.run, "func") as mock:
        TaskDir.run.func.__annotations__ = {"return": Dir}

        mock.return_value = None
        with pytest.raises(NotSupported):
            TaskDir().run()
        mock.assert_called_once()
        mock.reset_mock()

        o = TaskDir().run.output.temp()
        assert isinstance(o, Dir)
        o = FSOutput(o)
        assert isinstance(o, FSOutput)
        assert not isinstance(o, Dir)
        assert o._temp
        mock.return_value = o
        with pytest.warns(Inconsistent):
            TaskDir().run()
        TaskDir().remove()
        mock.assert_called_once()
        mock.reset_mock()

        o = TaskDir().run.output.temp() / "not_existing"
        assert isinstance(o, Dir)
        mock.return_value = o
        with pytest.raises(MissingOuput):
            TaskDir().run()
        mock.assert_called_once()
        mock.reset_mock()

        f: File = TaskDir().run.output.temp().joinpath("some.file")
        f.touch()
        assert f.is_file()
        mock.return_value = f
        with pytest.raises(IsADirectoryError):
            TaskDir().run()
        mock.assert_called_once()
        mock.reset_mock()


def test_output_Dir_temp(tmpdir4BaseTask):
    out: Output[Dir] = TaskDir().run.output
    with pytest.raises(ValueError):
        out.temp("foo")
    with pytest.raises(ValueError):
        out.temp(anything=True)

    tmp = out.temp()
    assert isinstance(tmp, Dir)
    assert isinstance(tmp._temp, TemporaryDirectory)

    sub = tmp / "inner.file"
    assert isinstance(sub, Dir)
    assert isinstance(sub._temp, TemporaryDirectory)
    assert tmp._temp is sub._temp

    parent = tmp.parent
    assert isinstance(parent, Dir)
    assert parent._temp is None


def test_output_Dir_remove(tmpdir4BaseTask):
    td = TaskDir()
    assert isinstance(td.run(), Dir)
    assert isinstance(td.run(), Dir)  # yes, again!
    assert isinstance(td.run.output.load(), Dir)

    d = td.run()
    assert d.is_dir()
    assert d.joinpath("foo.bar").is_file()
    td.run.output.remove()
    assert not d.exists()

    td.run.output.remove(missing_ok=True)
    with pytest.raises(FileNotFoundError):
        td.run.output.remove(missing_ok=False)
    with pytest.raises(FileNotFoundError):
        td.run.output.remove()


def test_bad_output():
    with catch_warnings(action="ignore", category=OwnParameters.CacheReset):

        class Foo(Task):
            def run(): ...

            output = None

        with pytest.raises(TypeError):
            Foo().run.output


def test_print(capsys):
    with patch.object(pprint, "_stream", sys.stdout):
        Task2().print()

        cap = capsys.readouterr()
        assert cap.err == ""
        assert cap.out.split("\n") == [
            "tests.test_task:Task2(dep=tests.test_task:Task1())",
            "  tests.test_task:Task1()",
            "",
        ]


def test_status(capsys, tmpdir4BaseTask):
    with patch.object(pprint, "_stream", sys.stdout):
        Task2().status()

        cap = capsys.readouterr()
        assert cap.err == ""
        assert cap.out.split("\n") == [
            "tests.test_task:Task2(dep=tests.test_task:Task1()): missing",
            "  tests.test_task:Task1(): missing",
            "",
        ]

        filler = "~".join(map(str, range(50)))

        Task2({(Task, "space_waster"): filler}).status()

        cap = capsys.readouterr()
        assert cap.err == ""
        assert cap.out.split("\n") == [
            "tests.test_task:Task2(",
            " dep=tests.test_task:Task1(...),",
            f" space_waster='{filler}'",
            "): missing",
            "  tests.test_task:Task1(",
            f"   space_waster='{filler}'",
            "  ): missing",
            "",
        ]


def test_remove(capsys, tmpdir4BaseTask):
    with patch.object(pprint, "_stream", sys.stdout):
        t2 = Task2()
        t2.run()

        t2.remove(0)
        cap = capsys.readouterr()
        assert cap.err == ""
        assert cap.out.split("\n") == [
            "tests.test_task:Task2(dep=tests.test_task:Task1()): removed",
            "",
        ]

        t2.remove()
        cap = capsys.readouterr()
        assert cap.err == ""
        assert cap.out.split("\n") == [
            "tests.test_task:Task2(dep=tests.test_task:Task1()): missing",
            "  tests.test_task:Task1(): removed",
            "",
        ]


def test_action_ordering():
    func = Mock()

    class Foo(ParaO):
        act1 = ValueAction[int, None](func)
        act2 = ValueAction[int, None](func)

    cli = CLI(entry_points=[Foo])

    [foo] = cli.run(["Foo", "--act1=1", "--act2=2"])
    func.assert_has_calls([call(foo, 1), call(foo, 2)])

    func.reset_mock()

    [foo] = cli.run(["Foo", "--act2=2", "--act1=1"])
    func.assert_has_calls([call(foo, 2), call(foo, 1)])


def test_ConcurrentRunner(tmpdir4BaseTask):
    runner = ConcurrentRunner(ThreadPoolExecutor(1))

    task = Task2()
    res = task.run()

    task.remove()
    assert task.run(runner=runner) == res
    assert task.run.done

    task.remove()
    with ConcurrentRunner.current(runner):
        assert task.run() == res
        assert task.run() == res  # yes twice
        assert task.run.done



from abc import ABC, abstractmethod
from concurrent.futures import Executor, wait
from typing import Callable, Iterable, Self, Type

from .action import BaseRecursiveAction, RecursiveAct
from .cast import Opaque
from .core import ParaO
from .misc import ContextValue


class PseudoOutput(Opaque):
    """Special type that must only be handled directly lest it cause an error."""


class BaseOutput[T](ABC):
    __slots__ = ("act",)

    def __init__(self, act: "RunAct[T]"):
        self.act = act

    @property
    @abstractmethod
    def exists(self) -> bool: ...

    @abstractmethod
    def load(self) -> T: ...

    @abstractmethod
    def dump(self, data: T | PseudoOutput) -> T: ...

    @abstractmethod
    def remove(self, missing_ok: bool = False) -> None: ...


class RunAct[T](RecursiveAct["RunAction[T]"]):
    __call__: Callable[[], T]

    def _func(
        self, sub: Iterable[Self], runner: "Runner[T] | None" = None, **kwargs
    ) -> T:
        if runner is None:
            runner = Runner.current()
        if runner is None:
            if self.done:
                return self.output.load()
            else:
                for s in sub:
                    if not s.done:
                        s(**kwargs)

                return self._make()
        else:
            return runner(self, sub, kwargs)

    def _make(self, output: bool = True):
        out = self.output.dump(self.action.func(self.instance))
        if output:
            return out

    __slots__ = ("output",)
    output: BaseOutput

    def __post_init__(self):
        super().__post_init__()
        output = self.action.output or self.instance.output
        if not (isinstance(output, type) and issubclass(output, BaseOutput)):
            raise TypeError(f"{output=} must by a BaseOutput subclass")
        object.__setattr__(self, "output", output(self))  # side-step "frozen"

    @property
    def done(self):
        return self.output.exists

    @property
    def _key(self):
        return self.instance, self.action


class RunAction[R](BaseRecursiveAction[R, []]):
    _act: Type[RunAct[R]] = RunAct
    func: Callable[[ParaO], R]
    __get__: Callable[..., RunAct[R]]
    output: Type[BaseOutput] | None = None


RunAction._peer_base = RunAction


class Runner(ABC):
    current = ContextValue["Runner | None"]("currentRunner", default=None)

    @abstractmethod
    def __call__[T](self, act: RunAct[T], sub: Iterable[RunAct], sub_kwargs: dict): ...


class ConcurrentRunner(Runner):
    def __init__(self, executor: Executor):
        self.executor = executor
        super().__init__()

    def __call__[T](self, act: RunAct[T], sub: Iterable[RunAct], sub_kwargs: dict):
        if act.done:
            return act.output.load()
        else:
            wait(self.executor.submit(s, **sub_kwargs) for s in sub if not s.done)
            return self.executor.submit(act._make).result()



from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from functools import lru_cache
from inspect import Parameter, signature
from operator import attrgetter
from typing import Any, Callable, Concatenate, Iterable, Self, Type

from .core import UNSET, AbstractDecoParam, ParaO, TypedAlias, Unset, Value, eager
from .misc import ContextValue

__all__ = ["SimpleAction", "ValueAction", "RecursiveAction"]


@lru_cache
def _method_1st_arg_annotation[T](
    func: Callable[Concatenate[Any, T, ...], Any],
) -> Type[T] | Unset:
    for i, param in enumerate(signature(func).parameters.values()):
        if i == 1:
            if param.kind in (
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
                Parameter.VAR_POSITIONAL,
            ):
                if param.annotation is Parameter.empty:
                    return UNSET
                else:
                    return param.annotation
            break
    return UNSET


@dataclass(slots=True, frozen=True)
class BaseAct[T, A: "BaseAction"](ABC):
    action: A
    instance: ParaO
    value: T
    position: int = 0

    def __post_init__(self):
        if self.trigger:
            Plan.add(self)

    @property
    def trigger(self):
        return self.value is not UNSET

    @property
    def name(self) -> str:
        return self.action._name(self.instance.__class__)

    @abstractmethod
    def __call__(self): ...


class BaseAction[T, R, **Ps](AbstractDecoParam[T, Callable[Concatenate[ParaO, Ps], R]]):
    significant = False
    _act: Type[BaseAct] = BaseAct

    TypedAlias.register(R, "return_type")

    def _type(self, cls, name):
        return self.type

    def _get(self, val, name, instance) -> BaseAct:
        pos = val.position if isinstance(val, Value) else 0
        val = super()._get(val, name, instance)
        return self._act(self, instance, val, pos)

    def _collect(self, expansion, instance):  # can't collect
        return False  # pragma: no cover

    __get__: Callable[..., BaseAct[T, Self]]


# simple variant
class SimpleAct[R](BaseAct[bool, "SimpleAction[R]"]):
    __slots__ = ()

    @property
    def trigger(self):
        return super().trigger and self.value

    def __call__(self) -> R:
        return self.action.func(self.instance)


class SimpleAction[R](BaseAction[bool, R, []]):
    _act = SimpleAct
    func: Callable[[ParaO], R]
    __get__: Callable[..., SimpleAct[R]]
    type = bool


# value variant
class ValueAct[T, R](BaseAct[T, "ValueAction[T, R]"]):
    __slots__ = ()

    def __call__(self, override: T | Unset = UNSET) -> R:
        value = self.value if override is UNSET else override
        if value is UNSET:
            return self.action.func(self.instance)
        else:
            return self.action.func(self.instance, value)


class ValueAction[T, R](BaseAction[T, R, [T]]):
    def _type(self, cls, name):
        typ = self.type
        if typ is UNSET:
            typ = _method_1st_arg_annotation(self.func)
        return typ

    _act = ValueAct
    func: Callable[[ParaO, T], R]
    __get__: Callable[..., ValueAct[T, R]]
    type: Type[T]


# recursive variant
class RecursiveAct[A: "RecursiveAction"](BaseAct[int | bool | None, A]):
    __slots__ = ()

    def _inner(self):
        name = self.name
        is_peer = self.action.__class__._is_peer
        for inner in self.instance.__inner__:
            if other := inner.__class__.__own_parameters__.get(name):
                if is_peer(other.__class__):
                    yield getattr(inner, name)

    def _func(self, sub: Iterable[Self], depth: int = 0, **kwargs):
        if not self.action.func(self.instance, depth):
            for s in sub:
                s(depth=depth + 1, **kwargs)

    def __call__(
        self, override: int | bool | None = None, *, _outer: int = None, **kwargs
    ):
        if override is None:
            val = self.value
            if val is UNSET:
                val = True if _outer is None else _outer
            elif self.trigger:  # pragma: no branch
                Plan.consume(self)
        else:
            val = override
        if val is False or val < 0:
            return

        return self._func(
            self._inner() if val else (),  # recusion elements, if (still) allowed
            _outer=val is True or val < 1 or val - 1,  # remaining recursion
            **kwargs,  # arbitrary other state, e.g. depth
        )


class BaseRecursiveAction[R, **Ps](BaseAction[int | bool | None, R, Ps]):
    _peer_base: type | None = None

    @classmethod
    def _is_peer(cls, other_cls: type):
        if base := cls._peer_base:
            return issubclass(other_cls, base)
        else:
            return other_cls is cls

    type = int | bool | None


class RecursiveAction(BaseRecursiveAction[bool, [int]]):
    func: Callable[[ParaO, int], bool]
    __get__: Callable[..., RecursiveAct]
    _act = RecursiveAct


class Plan(list[BaseAct]):
    current = ContextValue["Plan"]("currentPlan", default=None)
    _sorted: bool = False

    @classmethod
    def add(cls, act: BaseAct):
        if (curr := cls.current()) is not None:
            curr.append(act)
            curr._sorted = False

    @classmethod
    def consume(cls, act: BaseAct):
        if not (curr := cls.current()):
            return
        try:
            idx = curr.index(act)
        except ValueError:
            return
        name = act.name
        is_peer = act.action.__class__._is_peer
        if all(a.name == name and is_peer(a.action.__class__) for a in curr[:idx]):
            del curr[idx]
            return True

    @contextmanager
    def use(self, /, run: bool = False):
        with self.current(self), eager(True):
            yield
            if run:
                self.run()

    def sort(self):
        if not self._sorted:
            super().sort(key=attrgetter("position"))
            self._sorted = True

    def run(self):
        while self:
            self.sort()
            self.pop(0)()



from json import JSONDecodeError
import sys
from unittest import TestCase
from unittest.mock import patch
from parao.cli import (
    CLI,
    AmbigouusCandidate,
    CLIParser,
    MalformedCommandline,
    MultipleCandidates,
    NotAParaO,
    ParaONotFound,
    Sep,
    UnsupportedKeyType,
    UnusedCLIArguments,
    ValueMissing,
)
from parao.core import MissingParameterValue, ParaO, Param


class Outer1(ParaO):
    class Inner(ParaO):
        foo = Param[int](2)
        boo = Param[str]("inner")

    foo = Param[int](1)
    bar = Param[str]("outer")
    boo = Param[bool](False)


class Outer2(ParaO):
    class Inner(ParaO):
        pass

    class Inner2(ParaO):
        pass

    foo = Param[int]()
    bar = Param[str]()


class Outer3(ParaO):
    class Inner2(ParaO):
        pass


Outer3.__module__ += ".sub"
Outer3.Inner2.__module__ += ".sub"


class MultiWrap(ParaO):
    inner1 = Param[ParaO]()
    inner2 = Param[ParaO]()


plain_object = object()


class TestCLI(TestCase):
    def test_argv(self):
        argv = ["<script>", "Outer1"]
        with patch("sys.argv", argv):
            self.assertEqual(sys.argv, argv)
            self.assertIsInstance(CLI().run()[0], Outer1)

    def test_plain(self):
        cli = CLI()

        a, b = cli.run(["Outer1", "Outer1.Inner"])
        self.assertIsInstance(a, Outer1)
        self.assertIsInstance(b, Outer1.Inner)
        with self.assertRaises(ParaONotFound, msg="DoesNotExist"):
            cli.run(["DoesNotExist"])
        with self.assertWarns(AmbigouusCandidate):
            self.assertRaises(ParaONotFound, lambda: cli.run(["Inner"]))
        with self.assertWarns(MultipleCandidates):
            self.assertRaises(ParaONotFound, lambda: cli.run(["Inner2"]))
        self.assertRaises(
            ModuleNotFoundError, lambda: cli.run(["does_not_exist.Inner2"])
        )
        self.assertIsInstance(cli.run(["sub.Inner2"])[0], Outer3.Inner2)
        with self.assertRaises(NotAParaO):
            cli.run(["tests.test_cli:plain_object"])

    def test_params(self):
        cli = CLI()

        self.assertEqual(cli.run(["Outer1", "--foo", "123"])[0].foo, 123)
        self.assertEqual(cli.run(["Outer1", "--foo=123"])[0].foo, 123)
        self.assertEqual(cli.run(["Outer1", "--Outer1.foo=123"])[0].foo, 123)
        # various empties
        self.assertEqual(cli.run(["Outer1", "--bar="])[0].bar, "")
        self.assertEqual(cli.run(["Outer1", "--boo"])[0].boo, True)
        self.assertEqual(cli.run(["Outer1", "--boo", "--bar=b"])[0].boo, True)
        # class
        with self.assertRaises(ValueMissing):
            cli.run(["MultiWrap", "--inner1,__class__="])
        with self.assertRaises(ParaONotFound):
            cli.run(["MultiWrap", "--inner1,__class__=ThisDoesNotExist"])
        (wrap,) = cli.run(["MultiWrap", "--inner1,__class__=Outer1"])
        self.assertIsInstance(wrap.inner1, Outer1)
        self.assertIsInstance(wrap.inner2, ParaO)
        # json
        self.assertEqual(len(cli.run(["Outer1", "--foo;json", "[1,2,3]"])), 3)
        with self.assertRaises(ValueMissing):
            cli.run(["Outer1", "--foo;json="])
        with self.assertRaises(JSONDecodeError):
            cli.run(["Outer1", "--foo;json=]"])
        # with module
        self.assertEqual(cli.run(["Outer1", "--test_cli.Outer1.foo=123"])[0].foo, 123)
        self.assertEqual(cli.run(["Outer1", "--test_cli:Outer1.foo=123"])[0].foo, 123)
        with self.assertRaises(ModuleNotFoundError):
            cli.run(["Outer1", "--test_cli:bad.Outer1.foo=123"])
        self.assertEqual(cli.run(["Outer1", "--not_found.foo=123"])[0].foo, 1)
        self.assertEqual(
            cli.run(["Outer1", "--tests.test_cli:Outer1.foo=123"])[0].foo, 123
        )
        with self.assertRaises(MalformedCommandline):
            cli.run(["Outer1", "--tests.test_cli:=123"])
        with self.assertWarns(UnsupportedKeyType), self.assertRaises(TypeError):
            cli.run(["Outer1", "--tests.test_cli:plain_object=123"])
        # expansion
        self.assertEqual(len(cli.run(["Outer1", "--foo=1,2"])), 2)
        with self.assertRaises(ValueError):
            cli.run(["Outer1", "--bar=A,B", "--foo=1,x"])

    def test_prio(self):
        cli = CLI()
        self.assertEqual(cli.run(["Outer1", "-foo=9", "-foo=1"])[0].foo, 1)
        self.assertEqual(cli.run(["Outer1", "+foo=9", "-foo=1"])[0].foo, 9)
        self.assertEqual(cli.run(["Outer1", "-+foo=9", "-foo=1"])[0].foo, 9)
        self.assertEqual(cli.run(["Outer1", "-foo;prio:=9", "-foo=1"])[0].foo, 1)
        self.assertEqual(cli.run(["Outer1", "-foo;prio:1=9", "-foo=1"])[0].foo, 9)
        self.assertEqual(cli.run(["Outer1", "-foo;prio:1.1=9", "-foo=1"])[0].foo, 9)
        with patch.object(Outer1.foo, "min_prio", 2):
            self.assertEqual(cli.run(["Outer1", "-foo=3"])[0].foo, 1)
        with self.assertRaises(ValueError):
            cli.run(["Outer1", "-foo;prio:x=9"])

    def test_unused_arguments(self):
        cli = CLI()
        self.assertEqual(cli.run([]), [])
        cli.run(["", "Outer1", ""])
        with self.assertWarns(UnusedCLIArguments):
            cli.run(["--foo", "Outer1"])
        with self.assertWarns(UnusedCLIArguments):
            cli.run(["Outer1", "--", "--foo"])

    def test_errors(self):
        self.assertRaises(MissingParameterValue, lambda: CLI().run(["Outer2"]))
        self.assertRaises(
            MissingParameterValue, lambda: CLI().run(["Outer2", "-foo", "1"])
        )
        self.assertRaises(
            MissingParameterValue, lambda: CLI().run(["Outer2", "-foo", "1,2"])
        )


class TestSep(TestCase):
    def test(self):
        self.assertRaises(Sep.NeedValues, lambda: Sep(()))
        self.assertRaises(Sep.Overlap, lambda: Sep("x") << Sep("xy"))
        self.assertEqual(Sep(("foo", "bar")).regex.pattern, "(?:foo|bar)")
        self.assertEqual(Sep((*"foo", "bar")).regex.pattern, "(?:bar|[foo])")


class TestCLIParser(TestCase):
    def test(self):
        p = CLIParser(flag="=")
        self.assertFalse(p._flag_value_disjoint)
        self.assertEqual(
            p.argument("foo=json=val"),
            ([("foo", None)], {"json": None}, "val"),
        )
        self.assertEqual(
            p.argument("foo="),
            ([("foo", None)], {"": None}, None),
        )
        self.assertEqual(
            p.argument("foo"),
            ([("foo", None)], None, None),
        )

        class Sub(CLIParser):
            extra: int = 0

        Sub()
